# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypfopt', 'pypfopt.efficient_frontier']

package_data = \
{'': ['*']}

install_requires = \
['cvxpy>=1.1.10,<2.0.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=0.19',
 'scipy>=1.3,<2.0']

extras_require = \
{'optionals': ['matplotlib>=3.5.2,<4.0.0', 'scikit-learn>=1.1.1,<2.0.0']}

setup_kwargs = {
    'name': 'pyportfolioopt',
    'version': '1.5.3',
    'description': 'Financial portfolio optimization in python',
    'long_description': '<p align="center">\n    <img width=60% src="https://github.com/robertmartin8/PyPortfolioOpt/blob/master/media/logo_v1.png?raw=true">\n</p>\n\n<!-- buttons -->\n<p align="center">\n    <a href="https://www.python.org/">\n        <img src="https://img.shields.io/badge/python-v3-brightgreen.svg"\n            alt="python"></a> &nbsp;\n    <a href="https://pypi.org/project/PyPortfolioOpt/">\n        <img src="https://img.shields.io/badge/pypi-v1.5.3-brightgreen.svg"\n            alt="pypi"></a> &nbsp;\n    <a href="https://opensource.org/licenses/MIT">\n        <img src="https://img.shields.io/badge/license-MIT-brightgreen.svg"\n            alt="MIT license"></a> &nbsp;\n    <a href="https://github.com/robertmartin8/PyPortfolioOpt/actions">\n        <img src="https://github.com/robertmartin8/PyPortfolioOpt/workflows/pytest/badge.svg?event=push"\n            alt="build"></a> &nbsp;\n    <a href="https://pepy.tech/project/pyportfolioopt">\n        <img src="https://pepy.tech/badge/pyportfolioopt"\n            alt="downloads"></a> &nbsp;\n    <a href="https://mybinder.org/v2/gh/robertmartin8/pyportfolioopt/master/?filepath=cookbook">\n      <img src="https://mybinder.org/badge_logo.svg"\n            alt="binder"></a> &nbsp;\n</p>\n\n<!-- content -->\n\n**PyPortfolioOpt has recently been [published](https://joss.theoj.org/papers/10.21105/joss.03066) in the Journal of Open Source Software 🎉**\n\nPyPortfolioOpt is a library that implements portfolio optimization methods, including\nclassical mean-variance optimization techniques and Black-Litterman allocation, as well as more\nrecent developments in the field like shrinkage and Hierarchical Risk Parity.\n\nIt is **extensive** yet easily **extensible**, and can be useful for either a casual investors, or a professional looking for an easy prototyping tool. Whether you are a fundamentals-oriented investor who has identified a\nhandful of undervalued picks, or an algorithmic trader who has a basket of\nstrategies, PyPortfolioOpt can help you combine your alpha sources\nin a risk-efficient way.\n\n**Update as of May 2022**: I am starting fulltime work, so don’t have a huge amount of bandwidth for PyPortfolioOpt anymore. I will aim to keep the package funcitonal, but will not be adding new features. Please check out Dany Cajas\' [Riskfolio-lib](https://riskfolio-lib.readthedocs.io/en/latest/index.html) if you would like more advanced features!\n\nHead over to the **[documentation on ReadTheDocs](https://pyportfolioopt.readthedocs.io/en/latest/)** to get an in-depth look at the project, or check out the [cookbook](https://github.com/robertmartin8/PyPortfolioOpt/tree/master/cookbook) to see some examples showing the full process from downloading data to building a portfolio.\n\n<center>\n<img src="https://github.com/robertmartin8/PyPortfolioOpt/blob/master/media/conceptual_flowchart_v2.png?raw=true" style="width:70%;"/>\n</center>\n\n## Table of contents\n\n-   [Table of contents](#table-of-contents)\n-   [Getting started](#getting-started)\n    -   [For development](#for-development)\n-   [A quick example](#a-quick-example)\n-   [An overview of classical portfolio optimization methods](#an-overview-of-classical-portfolio-optimization-methods)\n-   [Features](#features)\n    -   [Expected returns](#expected-returns)\n    -   [Risk models (covariance)](#risk-models-covariance)\n    -   [Objective functions](#objective-functions)\n    -   [Adding constraints or different objectives](#adding-constraints-or-different-objectives)\n    -   [Black-Litterman allocation](#black-litterman-allocation)\n    -   [Other optimizers](#other-optimizers)\n-   [Advantages over existing implementations](#advantages-over-existing-implementations)\n-   [Project principles and design decisions](#project-principles-and-design-decisions)\n-   [Testing](#testing)\n-   [Citing PyPortfolioOpt](#citing-pyportfolioopt)\n-   [Contributing](#contributing)\n-   [Getting in touch](#getting-in-touch)\n\n## Getting started\n\nIf you would like to play with PyPortfolioOpt interactively in your browser, you may launch Binder [here](https://mybinder.org/v2/gh/robertmartin8/pyportfolioopt/master). It takes a\nwhile to set up, but it lets you try out the cookbook recipes without having to deal with all of the requirements.\n\n_Note: macOS users will need to install [Command Line Tools](https://osxdaily.com/2014/02/12/install-command-line-tools-mac-os-x/)._\n\n_Note: if you are on windows, you first need to installl C++. ([download](https://visualstudio.microsoft.com/thank-you-downloading-visual-studio/?sku=BuildTools&rel=16), [install instructions](https://drive.google.com/file/d/0B4GsMXCRaSSIOWpYQkstajlYZ0tPVkNQSElmTWh1dXFaYkJr/view))_\n\nThis project is available on PyPI, meaning that you can just:\n\n```bash\npip install PyPortfolioOpt\n```\n\n(you may need to follow separate installation instructions for [cvxopt](https://cvxopt.org/install/index.html#) and [cvxpy](https://www.cvxpy.org/install/))).\n\nHowever, it is best practice to use a dependency manager within a virtual environment.\nMy current recommendation is to get yourself set up with [poetry](https://github.com/sdispater/poetry) then just run\n\n```bash\npoetry add PyPortfolioOpt\n```\n\nOtherwise, clone/download the project and in the project directory run:\n\n```bash\npython setup.py install\n```\n\nPyPortfolioOpt supports Docker. Build your first container with `docker build -f docker/Dockerfile . -t pypfopt`. You can use the image to run tests or even launch a Jupyter server.\n\n```bash\n# iPython interpreter:\ndocker run -it pypfopt poetry run ipython\n\n# Jupyter notebook server:\ndocker run -it -p 8888:8888 pypfopt poetry run jupyter notebook --allow-root --no-browser --ip 0.0.0.0\n# click on http://127.0.0.1:8888/?token=xxx\n\n# Pytest\ndocker run -t pypfopt poetry run pytest\n\n# Bash\ndocker run -it pypfopt bash\n```\n\nFor more information, please read [this guide](https://docker-curriculum.com/#introduction).\n\n### For development\n\nIf you would like to make major changes to integrate this with your proprietary system, it probably makes sense to clone this repository and to just use the source code.\n\n```bash\ngit clone https://github.com/robertmartin8/PyPortfolioOpt\n```\n\nAlternatively, you could try:\n\n```bash\npip install -e git+https://github.com/robertmartin8/PyPortfolioOpt.git\n```\n\n## A quick example\n\nHere is an example on real life stock data, demonstrating how easy it is to find the long-only portfolio that maximises the Sharpe ratio (a measure of risk-adjusted returns).\n\n```python\nimport pandas as pd\nfrom pypfopt import EfficientFrontier\nfrom pypfopt import risk_models\nfrom pypfopt import expected_returns\n\n# Read in price data\ndf = pd.read_csv("tests/resources/stock_prices.csv", parse_dates=True, index_col="date")\n\n# Calculate expected returns and sample covariance\nmu = expected_returns.mean_historical_return(df)\nS = risk_models.sample_cov(df)\n\n# Optimize for maximal Sharpe ratio\nef = EfficientFrontier(mu, S)\nraw_weights = ef.max_sharpe()\ncleaned_weights = ef.clean_weights()\nef.save_weights_to_file("weights.csv")  # saves to file\nprint(cleaned_weights)\nef.portfolio_performance(verbose=True)\n```\n\nThis outputs the following weights:\n\n```txt\n{\'GOOG\': 0.03835,\n \'AAPL\': 0.0689,\n \'FB\': 0.20603,\n \'BABA\': 0.07315,\n \'AMZN\': 0.04033,\n \'GE\': 0.0,\n \'AMD\': 0.0,\n \'WMT\': 0.0,\n \'BAC\': 0.0,\n \'GM\': 0.0,\n \'T\': 0.0,\n \'UAA\': 0.0,\n \'SHLD\': 0.0,\n \'XOM\': 0.0,\n \'RRC\': 0.0,\n \'BBY\': 0.01324,\n \'MA\': 0.35349,\n \'PFE\': 0.1957,\n \'JPM\': 0.0,\n \'SBUX\': 0.01082}\n\nExpected annual return: 30.5%\nAnnual volatility: 22.2%\nSharpe Ratio: 1.28\n```\n\nThis is interesting but not useful in itself. However, PyPortfolioOpt provides a method which allows you to convert the above continuous weights to an actual allocation that you could buy. Just enter the most recent prices, and the desired portfolio size ($10,000 in this example):\n\n```python\nfrom pypfopt.discrete_allocation import DiscreteAllocation, get_latest_prices\n\n\nlatest_prices = get_latest_prices(df)\n\nda = DiscreteAllocation(weights, latest_prices, total_portfolio_value=10000)\nallocation, leftover = da.greedy_portfolio()\nprint("Discrete allocation:", allocation)\nprint("Funds remaining: ${:.2f}".format(leftover))\n```\n\n```txt\n12 out of 20 tickers were removed\nDiscrete allocation: {\'GOOG\': 1, \'AAPL\': 4, \'FB\': 12, \'BABA\': 4, \'BBY\': 2,\n                      \'MA\': 20, \'PFE\': 54, \'SBUX\': 1}\nFunds remaining: $11.89\n```\n\n_Disclaimer: nothing about this project constitues investment advice, and the author bears no responsibiltiy for your subsequent investment decisions. Please refer to the [license](https://github.com/robertmartin8/PyPortfolioOpt/blob/master/LICENSE.txt) for more information._\n\n## An overview of classical portfolio optimization methods\n\nHarry Markowitz\'s 1952 paper is the undeniable classic, which turned portfolio optimization from an art into a science. The key insight is that by combining assets with different expected returns and volatilities, one can decide on a mathematically optimal allocation which minimises the risk for a target return – the set of all such optimal portfolios is referred to as the **efficient frontier**.\n\n<center>\n<img src="https://github.com/robertmartin8/PyPortfolioOpt/blob/master/media/efficient_frontier_white.png?raw=true" style="width:60%;"/>\n</center>\n\nAlthough much development has been made in the subject, more than half a century later, Markowitz\'s core ideas are still fundamentally important and see daily use in many portfolio management firms.\nThe main drawback of mean-variance optimization is that the theoretical treatment requires knowledge of the expected returns and the future risk-characteristics (covariance) of the assets. Obviously, if we knew the expected returns of a stock life would be much easier, but the whole game is that stock returns are notoriously hard to forecast. As a substitute, we can derive estimates of the expected return and covariance based on historical data – though we do lose the theoretical guarantees provided by Markowitz, the closer our estimates are to the real values, the better our portfolio will be.\n\nThus this project provides four major sets of functionality (though of course they are intimately related)\n\n-   Estimates of expected returns\n-   Estimates of risk (i.e covariance of asset returns)\n-   Objective functions to be optimized\n-   Optimizers.\n\nA key design goal of PyPortfolioOpt is **modularity** – the user should be able to swap in their\ncomponents while still making use of the framework that PyPortfolioOpt provides.\n\n## Features\n\nIn this section, we detail some of PyPortfolioOpt\'s available functionality. More examples are offered in the Jupyter notebooks [here](https://github.com/robertmartin8/PyPortfolioOpt/tree/master/cookbook). Another good resource is the [tests](https://github.com/robertmartin8/PyPortfolioOpt/tree/master/tests).\n\nA far more comprehensive version of this can be found on [ReadTheDocs](https://pyportfolioopt.readthedocs.io/en/latest/), as well as possible extensions for more advanced users.\n\n### Expected returns\n\n-   Mean historical returns:\n    -   the simplest and most common approach, which states that the expected return of each asset is equal to the mean of its historical returns.\n    -   easily interpretable and very intuitive\n-   Exponentially weighted mean historical returns:\n    -   similar to mean historical returns, except it gives exponentially more weight to recent prices\n    -   it is likely the case that an asset\'s most recent returns hold more weight than returns from 10 years ago when it comes to estimating future returns.\n-   Capital Asset Pricing Model (CAPM):\n    -   a simple model to predict returns based on the beta to the market\n    -   this is used all over finance!\n\n### Risk models (covariance)\n\nThe covariance matrix encodes not just the volatility of an asset, but also how it correlated to other assets. This is important because in order to reap the benefits of diversification (and thus increase return per unit risk), the assets in the portfolio should be as uncorrelated as possible.\n\n-   Sample covariance matrix:\n    -   an unbiased estimate of the covariance matrix\n    -   relatively easy to compute\n    -   the de facto standard for many years\n    -   however, it has a high estimation error, which is particularly dangerous in mean-variance optimization because the optimizer is likely to give excess weight to these erroneous estimates.\n-   Semicovariance: a measure of risk that focuses on downside variation.\n-   Exponential covariance: an improvement over sample covariance that gives more weight to recent data\n-   Covariance shrinkage: techniques that involve combining the sample covariance matrix with a structured estimator, to reduce the effect of erroneous weights. PyPortfolioOpt provides wrappers around the efficient vectorised implementations provided by `sklearn.covariance`.\n    -   manual shrinkage\n    -   Ledoit Wolf shrinkage, which chooses an optimal shrinkage parameter. We offer three shrinkage targets: `constant_variance`, `single_factor`, and `constant_correlation`.\n    -   Oracle Approximating Shrinkage\n-   Minimum Covariance Determinant:\n    -   a robust estimate of the covariance\n    -   implemented in `sklearn.covariance`\n\n<p align="center">\n    <img width=60% src="https://github.com/robertmartin8/PyPortfolioOpt/blob/master/media/corrplot_white.png?raw=true">\n</p>\n\n(This plot was generated using `plotting.plot_covariance`)\n\n### Objective functions\n\n-   Maximum Sharpe ratio: this results in a _tangency portfolio_ because on a graph of returns vs risk, this portfolio corresponds to the tangent of the efficient frontier that has a y-intercept equal to the risk-free rate. This is the default option because it finds the optimal return per unit risk.\n-   Minimum volatility. This may be useful if you\'re trying to get an idea of how low the volatility _could_ be, but in practice it makes a lot more sense to me to use the portfolio that maximises the Sharpe ratio.\n-   Efficient return, a.k.a. the Markowitz portfolio, which minimises risk for a given target return – this was the main focus of Markowitz 1952\n-   Efficient risk: the Sharpe-maximising portfolio for a given target risk.\n-   Maximum quadratic utility. You can provide your own risk-aversion level and compute the appropriate portfolio.\n\n### Adding constraints or different objectives\n\n-   Long/short: by default all of the mean-variance optimization methods in PyPortfolioOpt are long-only, but they can be initialised to allow for short positions by changing the weight bounds:\n\n```python\nef = EfficientFrontier(mu, S, weight_bounds=(-1, 1))\n```\n\n-   Market neutrality: for the `efficient_risk` and `efficient_return` methods, PyPortfolioOpt provides an option to form a market-neutral portfolio (i.e weights sum to zero). This is not possible for the max Sharpe portfolio and the min volatility portfolio because in those cases because they are not invariant with respect to leverage. Market neutrality requires negative weights:\n\n```python\nef = EfficientFrontier(mu, S, weight_bounds=(-1, 1))\nef.efficient_return(target_return=0.2, market_neutral=True)\n```\n\n-   Minimum/maximum position size: it may be the case that you want no security to form more than 10% of your portfolio. This is easy to encode:\n\n```python\nef = EfficientFrontier(mu, S, weight_bounds=(0, 0.1))\n```\n\nOne issue with mean-variance optimization is that it leads to many zero-weights. While these are\n"optimal" in-sample, there is a large body of research showing that this characteristic leads\nmean-variance portfolios to underperform out-of-sample. To that end, I have introduced an\nobjective function that can reduce the number of negligible weights for any of the objective functions. Essentially, it adds a penalty (parameterised by `gamma`) on small weights, with a term that looks just like L2 regularisation in machine learning. It may be necessary to try several `gamma` values to achieve the desired number of non-negligible weights. For the test portfolio of 20 securities, `gamma ~ 1` is sufficient\n\n```python\nef = EfficientFrontier(mu, S)\nef.add_objective(objective_functions.L2_reg, gamma=1)\nef.max_sharpe()\n```\n\n### Black-Litterman allocation\n\nAs of v0.5.0, we now support Black-Litterman asset allocation, which allows you to combine\na prior estimate of returns (e.g the market-implied returns) with your own views to form a\nposterior estimate. This results in much better estimates of expected returns than just using\nthe mean historical return. Check out the [docs](https://pyportfolioopt.readthedocs.io/en/latest/BlackLitterman.html) for a discussion of the theory, as well as advice\non formatting inputs.\n\n```python\nS = risk_models.sample_cov(df)\nviewdict = {"AAPL": 0.20, "BBY": -0.30, "BAC": 0, "SBUX": -0.2, "T": 0.131321}\nbl = BlackLittermanModel(S, pi="equal", absolute_views=viewdict, omega="default")\nrets = bl.bl_returns()\n\nef = EfficientFrontier(rets, S)\nef.max_sharpe()\n```\n\n### Other optimizers\n\nThe features above mostly pertain to solving mean-variance optimization problems via quadratic programming (though this is taken care of by `cvxpy`). However, we offer different optimizers as well:\n\n-   Mean-semivariance optimization\n-   Mean-CVaR optimization\n-   Hierarchical Risk Parity, using clustering algorithms to choose uncorrelated assets\n-   Markowitz\'s critical line algorithm (CLA)\n\nPlease refer to the [documentation](https://pyportfolioopt.readthedocs.io/en/latest/OtherOptimizers.html) for more.\n\n## Advantages over existing implementations\n\n-   Includes both classical methods (Markowitz 1952 and Black-Litterman), suggested best practices\n    (e.g covariance shrinkage), along with many recent developments and novel\n    features, like L2 regularisation, shrunk covariance, hierarchical risk parity.\n-   Native support for pandas dataframes: easily input your daily prices data.\n-   Extensive practical tests, which use real-life data.\n-   Easy to combine with your proprietary strategies and models.\n-   Robust to missing data, and price-series of different lengths (e.g FB data\n    only goes back to 2012 whereas AAPL data goes back to 1980).\n\n## Project principles and design decisions\n\n-   It should be easy to swap out individual components of the optimization process\n    with the user\'s proprietary improvements.\n-   Usability is everything: it is better to be self-explanatory than consistent.\n-   There is no point in portfolio optimization unless it can be practically\n    applied to real asset prices.\n-   Everything that has been implemented should be tested.\n-   Inline documentation is good: dedicated (separate) documentation is better.\n    The two are not mutually exclusive.\n-   Formatting should never get in the way of coding: because of this,\n    I have deferred **all** formatting decisions to [Black](https://github.com/ambv/black).\n\n## Testing\n\nTests are written in pytest (much more intuitive than `unittest` and the variants in my opinion), and I have tried to ensure close to 100% coverage. Run the tests by navigating to the package directory and simply running `pytest` on the command line.\n\nPyPortfolioOpt provides a test dataset of daily returns for 20 tickers:\n\n```python\n[\'GOOG\', \'AAPL\', \'FB\', \'BABA\', \'AMZN\', \'GE\', \'AMD\', \'WMT\', \'BAC\', \'GM\',\n\'T\', \'UAA\', \'SHLD\', \'XOM\', \'RRC\', \'BBY\', \'MA\', \'PFE\', \'JPM\', \'SBUX\']\n```\n\nThese tickers have been informally selected to meet several criteria:\n\n-   reasonably liquid\n-   different performances and volatilities\n-   different amounts of data to test robustness\n\nCurrently, the tests have not explored all of the edge cases and combinations\nof objective functions and parameters. However, each method and parameter has\nbeen tested to work as intended.\n\n## Citing PyPortfolioOpt\n\nIf you use PyPortfolioOpt for published work, please cite the [JOSS paper](https://joss.theoj.org/papers/10.21105/joss.03066).\n\nCitation string:\n\n```text\nMartin, R. A., (2021). PyPortfolioOpt: portfolio optimization in Python. Journal of Open Source Software, 6(61), 3066, https://doi.org/10.21105/joss.03066\n```\n\nBibTex::\n\n```bibtex\n@article{Martin2021,\n  doi = {10.21105/joss.03066},\n  url = {https://doi.org/10.21105/joss.03066},\n  year = {2021},\n  publisher = {The Open Journal},\n  volume = {6},\n  number = {61},\n  pages = {3066},\n  author = {Robert Andrew Martin},\n  title = {PyPortfolioOpt: portfolio optimization in Python},\n  journal = {Journal of Open Source Software}\n}\n```\n\n## Contributing\n\nContributions are _most welcome_. Have a look at the [Contribution Guide](https://github.com/robertmartin8/PyPortfolioOpt/blob/master/CONTRIBUTING.md) for more.\n\nI\'d like to thank all of the people who have contributed to PyPortfolioOpt since its release in 2018.\nSpecial shout-outs to:\n\n-   Philipp Schiele\n-   Carl Peasnell\n-   Felipe Schneider\n-   Dingyuan Wang\n-   Pat Newell\n-   Aditya Bhutra\n-   Thomas Schmelzer\n-   Rich Caputo\n-   Nicolas Knudde\n\n## Getting in touch\n\nIf you are having a problem with PyPortfolioOpt, please raise a GitHub issue. For anything else, you can reach me at:\n\n<center>\n<img src="https://github.com/robertmartin8/ReasonableDeviations/blob/gh-pages/assets/images/contact.png?raw=true" style="width:60%;"/>\n</center>\n',
    'author': 'Robert Andrew Martin',
    'author_email': 'martin.robertandrew@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/robertmartin8/PyPortfolioOpt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
The ``hierarchical_portfolio`` module seeks to implement one of the recent advances in
portfolio optimization – the application of hierarchical clustering models in allocation.

All of the hierarchical classes have a similar API to ``EfficientFrontier``, though since
many hierarchical models currently don't support different objectives, the actual allocation
happens with a call to `optimize()`.

Currently implemented:

- ``HRPOpt`` implements the Hierarchical Risk Parity (HRP) portfolio. Code reproduced with
  permission from Marcos Lopez de Prado (2016).
"""

import collections
import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as ssd

from . import base_optimizer, risk_models


class HRPOpt(base_optimizer.BaseOptimizer):
    """
    A HRPOpt object (inheriting from BaseOptimizer) constructs a hierarchical
    risk parity portfolio.

    Instance variables:

    - Inputs

        - ``n_assets`` - int
        - ``tickers`` - str list
        - ``returns`` - pd.DataFrame

    - Output:

        - ``weights`` - np.ndarray
        - ``clusters`` - linkage matrix corresponding to clustered assets.

    Public methods:

    - ``optimize()`` calculates weights using HRP
    - ``portfolio_performance()`` calculates the expected return, volatility and Sharpe ratio for
      the optimized portfolio.
    - ``set_weights()`` creates self.weights (np.ndarray) from a weights dict
    - ``clean_weights()`` rounds the weights and clips near-zeros.
    - ``save_weights_to_file()`` saves the weights to csv, json, or txt.
    """

    def __init__(self, returns=None, cov_matrix=None):
        """
        :param returns: asset historical returns
        :type returns: pd.DataFrame
        :param cov_matrix: covariance of asset returns
        :type cov_matrix: pd.DataFrame.
        :raises TypeError: if ``returns`` is not a dataframe
        """
        if returns is None and cov_matrix is None:
            raise ValueError("Either returns or cov_matrix must be provided")

        if returns is not None and not isinstance(returns, pd.DataFrame):
            raise TypeError("returns are not a dataframe")

        self.returns = returns
        self.cov_matrix = cov_matrix
        self.clusters = None

        if returns is None:
            tickers = list(cov_matrix.columns)
        else:
            tickers = list(returns.columns)
        super().__init__(len(tickers), tickers)

    @staticmethod
    def _get_cluster_var(cov, cluster_items):
        """
        Compute the variance per cluster

        :param cov: covariance matrix
        :type cov: np.ndarray
        :param cluster_items: tickers in the cluster
        :type cluster_items: list
        :return: the variance per cluster
        :rtype: float
        """
        # Compute variance per cluster
        cov_slice = cov.loc[cluster_items, cluster_items]
        weights = 1 / np.diag(cov_slice)  # Inverse variance weights
        weights /= weights.sum()
        return np.linalg.multi_dot((weights, cov_slice, weights))

    @staticmethod
    def _get_quasi_diag(link):
        """
        Sort clustered items by distance

        :param link: linkage matrix after clustering
        :type link: np.ndarray
        :return: sorted list of indices
        :rtype: list
        """
        return sch.to_tree(link, rd=False).pre_order()

    @staticmethod
    def _raw_hrp_allocation(cov, ordered_tickers):
        """
        Given the clusters, compute the portfolio that minimises risk by
        recursively traversing the hierarchical tree from the top.

        :param cov: covariance matrix
        :type cov: np.ndarray
        :param ordered_tickers: list of tickers ordered by distance
        :type ordered_tickers: str list
        :return: raw portfolio weights
        :rtype: pd.Series
        """
        w = pd.Series(1, index=ordered_tickers)
        cluster_items = [ordered_tickers]  # initialize all items in one cluster

        while len(cluster_items) > 0:
            cluster_items = [
                i[j:k]
                for i in cluster_items
                for j, k in ((0, len(i) // 2), (len(i) // 2, len(i)))
                if len(i) > 1
            ]  # bi-section
            # For each pair, optimize locally.
            for i in range(0, len(cluster_items), 2):
                first_cluster = cluster_items[i]
                second_cluster = cluster_items[i + 1]
                # Form the inverse variance portfolio for this pair
                first_variance = HRPOpt._get_cluster_var(cov, first_cluster)
                second_variance = HRPOpt._get_cluster_var(cov, second_cluster)
                alpha = 1 - first_variance / (first_variance + second_variance)
                w[first_cluster] *= alpha  # weight 1
                w[second_cluster] *= 1 - alpha  # weight 2
        return w

    def optimize(self, linkage_method="single"):
        """
        Construct a hierarchical risk parity portfolio, using Scipy hierarchical clustering
        (see `here <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html>`_)

        :param linkage_method: which scipy linkage method to use
        :type linkage_method: str
        :return: weights for the HRP portfolio
        :rtype: OrderedDict
        """
        if linkage_method not in sch._LINKAGE_METHODS:
            raise ValueError("linkage_method must be one recognised by scipy")

        if self.returns is None:
            cov = self.cov_matrix
            corr = risk_models.cov_to_corr(self.cov_matrix).round(6)
        else:
            corr, cov = self.returns.corr(), self.returns.cov()

        # Compute distance matrix, with ClusterWarning fix as
        # per https://stackoverflow.com/questions/18952587/

        # this can avoid some nasty floating point issues
        matrix = np.sqrt(np.clip((1.0 - corr) / 2.0, a_min=0.0, a_max=1.0))
        dist = ssd.squareform(matrix, checks=False)

        self.clusters = sch.linkage(dist, linkage_method)
        sort_ix = HRPOpt._get_quasi_diag(self.clusters)
        ordered_tickers = corr.index[sort_ix].tolist()
        hrp = HRPOpt._raw_hrp_allocation(cov, ordered_tickers)
        weights = collections.OrderedDict(hrp.sort_index())
        self.set_weights(weights)
        return weights

    def portfolio_performance(self, verbose=False, risk_free_rate=0.02, frequency=252):
        """
        After optimising, calculate (and optionally print) the performance of the optimal
        portfolio. Currently calculates expected return, volatility, and the Sharpe ratio
        assuming returns are daily

        :param verbose: whether performance should be printed, defaults to False
        :type verbose: bool, optional
        :param risk_free_rate: risk-free rate of borrowing/lending, defaults to 0.02.
                               The period of the risk-free rate should correspond to the
                               frequency of expected returns.
        :type risk_free_rate: float, optional
        :param frequency: number of time periods in a year, defaults to 252 (the number
                            of trading days in a year)
        :type frequency: int, optional
        :raises ValueError: if weights have not been calculated yet
        :return: expected return, volatility, Sharpe ratio.
        :rtype: (float, float, float)
        """
        if self.returns is None:
            cov = self.cov_matrix
            mu = None
        else:
            cov = self.returns.cov() * frequency
            mu = self.returns.mean() * frequency

        return base_optimizer.portfolio_performance(
            self.weights, mu, cov, verbose, risk_free_rate
        )



"""
The ``black_litterman`` module houses the BlackLittermanModel class, which
generates posterior estimates of expected returns given a prior estimate and user-supplied
views. In addition, two utility functions are defined, which calculate:

- market-implied prior estimate of returns
- market-implied risk-aversion parameter
"""
import sys
import warnings
import numpy as np
import pandas as pd
from . import base_optimizer


def market_implied_prior_returns(
    market_caps, risk_aversion, cov_matrix, risk_free_rate=0.02
):
    r"""
    Compute the prior estimate of returns implied by the market weights.
    In other words, given each asset's contribution to the risk of the market
    portfolio, how much are we expecting to be compensated?

    .. math::

        \Pi = \delta \Sigma w_{mkt}

    :param market_caps: market capitalisations of all assets
    :type market_caps: {ticker: cap} dict or pd.Series
    :param risk_aversion: risk aversion parameter
    :type risk_aversion: positive float
    :param cov_matrix: covariance matrix of asset returns
    :type cov_matrix: pd.DataFrame
    :param risk_free_rate: risk-free rate of borrowing/lending, defaults to 0.02.
                           You should use the appropriate time period, corresponding
                           to the covariance matrix.
    :type risk_free_rate: float, optional
    :return: prior estimate of returns as implied by the market caps
    :rtype: pd.Series
    """
    if not isinstance(cov_matrix, pd.DataFrame):
        warnings.warn(
            "If cov_matrix is not a dataframe, market cap index must be aligned to cov_matrix",
            RuntimeWarning,
        )
    mcaps = pd.Series(market_caps)
    mkt_weights = mcaps / mcaps.sum()
    # Pi is excess returns so must add risk_free_rate to get return.
    return risk_aversion * cov_matrix.dot(mkt_weights) + risk_free_rate


def market_implied_risk_aversion(market_prices, frequency=252, risk_free_rate=0.02):
    r"""
    Calculate the market-implied risk-aversion parameter (i.e market price of risk)
    based on market prices. For example, if the market has excess returns of 10% a year
    with 5% variance, the risk-aversion parameter is 2, i.e you have to be compensated 2x
    the variance.

    .. math::

        \delta = \frac{R - R_f}{\sigma^2}

    :param market_prices: the (daily) prices of the market portfolio, e.g SPY.
    :type market_prices: pd.Series with DatetimeIndex.
    :param frequency: number of time periods in a year, defaults to 252 (the number
                      of trading days in a year)
    :type frequency: int, optional
    :param risk_free_rate: risk-free rate of borrowing/lending, defaults to 0.02.
                            The period of the risk-free rate should correspond to the
                            frequency of expected returns.
    :type risk_free_rate: float, optional
    :raises TypeError: if market_prices cannot be parsed
    :return: market-implied risk aversion
    :rtype: float
    """
    if not isinstance(market_prices, (pd.Series, pd.DataFrame)):
        raise TypeError("Please format market_prices as a pd.Series")
    rets = market_prices.pct_change().dropna()
    r = rets.mean() * frequency
    var = rets.var() * frequency
    return (r - risk_free_rate) / var


class BlackLittermanModel(base_optimizer.BaseOptimizer):

    """
    A BlackLittermanModel object (inheriting from BaseOptimizer) contains requires
    a specific input format, specifying the prior, the views, the uncertainty in views,
    and a picking matrix to map views to the asset universe. We can then compute
    posterior estimates of returns and covariance. Helper methods have been provided
    to supply defaults where possible.

    Instance variables:

    - Inputs:

        - ``cov_matrix`` - np.ndarray
        - ``n_assets`` - int
        - ``tickers`` - str list
        - ``Q`` - np.ndarray
        - ``P`` - np.ndarray
        - ``pi`` - np.ndarray
        - ``omega`` - np.ndarray
        - ``tau`` - float

    - Output:

        - ``posterior_rets`` - pd.Series
        - ``posterior_cov`` - pd.DataFrame
        - ``weights`` - np.ndarray

    Public methods:

    - ``default_omega()`` - view uncertainty proportional to asset variance
    - ``idzorek_method()`` - convert views specified as percentages into BL uncertainties
    - ``bl_returns()`` - posterior estimate of returns
    - ``bl_cov()`` - posterior estimate of covariance
    - ``bl_weights()`` - weights implied by posterior returns
    - ``portfolio_performance()`` calculates the expected return, volatility
      and Sharpe ratio for the allocated portfolio.
    - ``set_weights()`` creates self.weights (np.ndarray) from a weights dict
    - ``clean_weights()`` rounds the weights and clips near-zeros.
    - ``save_weights_to_file()`` saves the weights to csv, json, or txt.
    """

    def __init__(
        self,
        cov_matrix,
        pi=None,
        absolute_views=None,
        Q=None,
        P=None,
        omega=None,
        view_confidences=None,
        tau=0.05,
        risk_aversion=1,
        **kwargs
    ):
        """
        :param cov_matrix: NxN covariance matrix of returns
        :type cov_matrix: pd.DataFrame or np.ndarray
        :param pi: Nx1 prior estimate of returns, defaults to None.
                   If pi="market", calculate a market-implied prior (requires market_caps
                   to be passed).
                   If pi="equal", use an equal-weighted prior.
        :type pi: np.ndarray, pd.Series, optional
        :param absolute_views: a colleciton of K absolute views on a subset of assets,
                               defaults to None. If this is provided, we do not need P, Q.
        :type absolute_views: pd.Series or dict, optional
        :param Q: Kx1 views vector, defaults to None
        :type Q: np.ndarray or pd.DataFrame, optional
        :param P: KxN picking matrix, defaults to None
        :type P: np.ndarray or pd.DataFrame, optional
        :param omega: KxK view uncertainty matrix (diagonal), defaults to None
                      Can instead pass "idzorek" to use Idzorek's method (requires
                      you to pass view_confidences). If omega="default" or None,
                      we set the uncertainty proportional to the variance.
        :type omega: np.ndarray or Pd.DataFrame, or string, optional
        :param view_confidences: Kx1 vector of percentage view confidences (between 0 and 1),
                                required to compute omega via Idzorek's method.
        :type view_confidences: np.ndarray, pd.Series, list, optional
        :param tau: the weight-on-views scalar (default is 0.05)
        :type tau: float, optional
        :param risk_aversion: risk aversion parameter, defaults to 1
        :type risk_aversion: positive float, optional
        :param market_caps: (kwarg) market caps for the assets, required if pi="market"
        :type market_caps: np.ndarray, pd.Series, optional
        :param risk_free_rate: (kwarg) risk_free_rate is needed in some methods
        :type risk_free_rate: float, defaults to 0.02
        """
        if sys.version_info[1] == 5:  # pragma: no cover
            warnings.warn(
                "When using python 3.5 you must explicitly construct the Black-Litterman inputs"
            )

        # Keep raw dataframes
        self._raw_cov_matrix = cov_matrix

        #  Initialise base optimizer
        if isinstance(cov_matrix, np.ndarray):
            self.cov_matrix = cov_matrix
            super().__init__(len(cov_matrix), list(range(len(cov_matrix))))
        else:
            self.cov_matrix = cov_matrix.values
            super().__init__(len(cov_matrix), cov_matrix.columns)

        #  Sanitise inputs
        if absolute_views is not None:
            self.Q, self.P = self._parse_views(absolute_views)
        else:
            self._set_Q_P(Q, P)
        self._set_risk_aversion(risk_aversion)
        self._set_pi(pi, **kwargs)
        self._set_tau(tau)
        # Make sure all dimensions work
        self._check_attribute_dimensions()

        self._set_omega(omega, view_confidences)

        # Private intermediaries
        self._tau_sigma_P = None
        self._A = None

        self.posterior_rets = None
        self.posterior_cov = None

    def _parse_views(self, absolute_views):
        """
        Given a collection (dict or series) of absolute views, construct
        the appropriate views vector and picking matrix. The views must
        be a subset of the tickers in the covariance matrix.

        {"AAPL": 0.20, "GOOG": 0.12, "XOM": -0.30}

        :param absolute_views: absolute views on asset performances
        :type absolute_views: dict, pd.Series
        """
        if not isinstance(absolute_views, (dict, pd.Series)):
            raise TypeError("views should be a dict or pd.Series")
        # Coerce to series
        views = pd.Series(absolute_views)
        k = len(views)

        Q = np.zeros((k, 1))
        P = np.zeros((k, self.n_assets))

        for i, view_ticker in enumerate(views.keys()):
            try:
                Q[i] = views[view_ticker]
                P[i, list(self.tickers).index(view_ticker)] = 1
            except ValueError:
                #  Could make this smarter by just skipping
                raise ValueError("Providing a view on an asset not in the universe")
        return Q, P

    def _set_Q_P(self, Q, P):
        if isinstance(Q, (pd.Series, pd.DataFrame)):
            self.Q = Q.values.reshape(-1, 1)
        elif isinstance(Q, np.ndarray):
            self.Q = Q.reshape(-1, 1)
        else:
            raise TypeError("Q must be an array or dataframe")

        if isinstance(P, pd.DataFrame):
            self.P = P.values
        elif isinstance(P, np.ndarray):
            self.P = P
        elif len(self.Q) == self.n_assets:
            # If a view on every asset is provided, P defaults
            # to the identity matrix.
            self.P = np.eye(self.n_assets)
        else:
            raise TypeError("P must be an array or dataframe")

    def _set_pi(self, pi, **kwargs):
        if pi is None:
            warnings.warn("Running Black-Litterman with no prior.")
            self.pi = np.zeros((self.n_assets, 1))
        elif isinstance(pi, (pd.Series, pd.DataFrame)):
            self.pi = pi.values.reshape(-1, 1)
        elif isinstance(pi, np.ndarray):
            self.pi = pi.reshape(-1, 1)
        elif pi == "market":
            if "market_caps" not in kwargs:
                raise ValueError(
                    "Please pass a series/array of market caps via the market_caps keyword argument"
                )
            market_caps = kwargs.get("market_caps")
            risk_free_rate = kwargs.get("risk_free_rate", 0)

            market_prior = market_implied_prior_returns(
                market_caps, self.risk_aversion, self._raw_cov_matrix, risk_free_rate
            )
            self.pi = market_prior.values.reshape(-1, 1)
        elif pi == "equal":
            self.pi = np.ones((self.n_assets, 1)) / self.n_assets
        else:
            raise TypeError("pi must be an array or series")

    def _set_tau(self, tau):
        if tau <= 0 or tau > 1:
            raise ValueError("tau should be between 0 and 1")
        self.tau = tau

    def _set_risk_aversion(self, risk_aversion):
        if risk_aversion <= 0:
            raise ValueError("risk_aversion should be a positive float")
        self.risk_aversion = risk_aversion

    def _set_omega(self, omega, view_confidences):
        if isinstance(omega, pd.DataFrame):
            self.omega = omega.values
        elif isinstance(omega, np.ndarray):
            self.omega = omega
        elif omega == "idzorek":
            if view_confidences is None:
                raise ValueError(
                    "To use Idzorek's method, please supply a vector of percentage "
                    "confidence levels for each view."
                )
            if not isinstance(view_confidences, np.ndarray):
                try:
                    view_confidences = np.array(view_confidences).reshape(-1, 1)
                    assert view_confidences.shape[0] == self.Q.shape[0]
                    assert np.issubdtype(view_confidences.dtype, np.number)
                except AssertionError:
                    raise ValueError(
                        "view_confidences should be a numpy 1D array or vector with the same length "
                        "as the number of views."
                    )

            self.omega = BlackLittermanModel.idzorek_method(
                view_confidences,
                self.cov_matrix,
                self.pi,
                self.Q,
                self.P,
                self.tau,
                self.risk_aversion,
            )
        elif omega is None or omega == "default":
            self.omega = BlackLittermanModel.default_omega(
                self.cov_matrix, self.P, self.tau
            )
        else:
            raise TypeError("self.omega must be a square array, dataframe, or string")

        K = len(self.Q)
        assert self.omega.shape == (K, K), "omega must have dimensions KxK"

    def _check_attribute_dimensions(self):
        """
        Helper method to ensure that all of the attributes created by the initialiser
        have the correct dimensions, to avoid linear algebra errors later on.

        :raises ValueError: if there are incorrect dimensions.
        """
        N = self.n_assets
        K = len(self.Q)
        assert self.pi.shape == (N, 1), "pi must have dimensions Nx1"
        assert self.P.shape == (K, N), "P must have dimensions KxN"
        assert self.cov_matrix.shape == (N, N), "cov_matrix must have shape NxN"

    @staticmethod
    def default_omega(cov_matrix, P, tau):
        """
        If the uncertainty matrix omega is not provided, we calculate using the method of
        He and Litterman (1999), such that the ratio omega/tau is proportional to the
        variance of the view portfolio.

        :return: KxK diagonal uncertainty matrix
        :rtype: np.ndarray
        """
        return np.diag(np.diag(tau * P @ cov_matrix @ P.T))

    @staticmethod
    def idzorek_method(view_confidences, cov_matrix, pi, Q, P, tau, risk_aversion=1):
        """
        Use Idzorek's method to create the uncertainty matrix given user-specified
        percentage confidences. We use the closed-form solution described by
        Jay Walters in The Black-Litterman Model in Detail (2014).

        :param view_confidences: Kx1 vector of percentage view confidences (between 0 and 1),
                                required to compute omega via Idzorek's method.
        :type view_confidences: np.ndarray, pd.Series, list,, optional
        :return: KxK diagonal uncertainty matrix
        :rtype: np.ndarray
        """
        view_omegas = []
        for view_idx in range(len(Q)):
            conf = view_confidences[view_idx]

            if conf < 0 or conf > 1:
                raise ValueError("View confidences must be between 0 and 1")

            # Special handler to avoid dividing by zero.
            # If zero conf, return very big number as uncertainty
            if conf == 0:
                view_omegas.append(1e6)
                continue

            P_view = P[view_idx].reshape(1, -1)
            alpha = (1 - conf) / conf  # formula (44)
            omega = tau * alpha * P_view @ cov_matrix @ P_view.T  # formula (41)
            view_omegas.append(omega.item())

        return np.diag(view_omegas)

    def bl_returns(self):
        """
        Calculate the posterior estimate of the returns vector,
        given views on some assets.

        :return: posterior returns vector
        :rtype: pd.Series
        """

        if self._tau_sigma_P is None:
            self._tau_sigma_P = self.tau * self.cov_matrix @ self.P.T

        # Solve the linear system Ax = b to avoid inversion
        if self._A is None:
            self._A = (self.P @ self._tau_sigma_P) + self.omega
        b = self.Q - self.P @ self.pi
        post_rets = self.pi + self._tau_sigma_P @ np.linalg.solve(self._A, b)
        return pd.Series(post_rets.flatten(), index=self.tickers)

    def bl_cov(self):
        """
        Calculate the posterior estimate of the covariance matrix,
        given views on some assets. Based on He and Litterman (2002).
        It is assumed that omega is diagonal. If this is not the case,
        please manually set omega_inv.

        :return: posterior covariance matrix
        :rtype: pd.DataFrame
        """
        if self._tau_sigma_P is None:
            self._tau_sigma_P = self.tau * self.cov_matrix @ self.P.T
        if self._A is None:
            self._A = (self.P @ self._tau_sigma_P) + self.omega

        b = self._tau_sigma_P.T
        M = self.tau * self.cov_matrix - self._tau_sigma_P @ np.linalg.solve(self._A, b)
        posterior_cov = self.cov_matrix + M
        return pd.DataFrame(posterior_cov, index=self.tickers, columns=self.tickers)

    def bl_weights(self, risk_aversion=None):
        r"""
        Compute the weights implied by the posterior returns, given the
        market price of risk. Technically this can be applied to any
        estimate of the expected returns, and is in fact a special case
        of mean-variance optimization

        .. math::

            w = (\delta \Sigma)^{-1} E(R)

        :param risk_aversion: risk aversion parameter, defaults to 1
        :type risk_aversion: positive float, optional
        :return: asset weights implied by returns
        :rtype: OrderedDict
        """
        if risk_aversion is None:
            risk_aversion = self.risk_aversion

        self.posterior_rets = self.bl_returns()
        A = risk_aversion * self.cov_matrix
        b = self.posterior_rets
        raw_weights = np.linalg.solve(A, b)
        self.weights = raw_weights / raw_weights.sum()
        return self._make_output_weights()

    def optimize(self, risk_aversion=None):
        """
        Alias for bl_weights for consistency with other methods.
        """
        return self.bl_weights(risk_aversion)

    def portfolio_performance(self, verbose=False, risk_free_rate=0.02):
        """
        After optimising, calculate (and optionally print) the performance of the optimal
        portfolio. Currently calculates expected return, volatility, and the Sharpe ratio.
        This method uses the BL posterior returns and covariance matrix.

        :param verbose: whether performance should be printed, defaults to False
        :type verbose: bool, optional
        :param risk_free_rate: risk-free rate of borrowing/lending, defaults to 0.02.
                               The period of the risk-free rate should correspond to the
                               frequency of expected returns.
        :type risk_free_rate: float, optional
        :raises ValueError: if weights have not been calcualted yet
        :return: expected return, volatility, Sharpe ratio.
        :rtype: (float, float, float)
        """
        if self.posterior_cov is None:
            self.posterior_cov = self.bl_cov()
        return base_optimizer.portfolio_performance(
            self.weights,
            self.posterior_rets,
            self.posterior_cov,
            verbose,
            risk_free_rate,
        )
